@@ -566,6 +566,18 @@ void SymbolTable::tentativeDefs(std::vector<const char*>& tents)
 }
 
 
+void SymbolTable::mustPreserveForBitcode(std::unordered_set<const char*>& syms)
+{
+	// return all names in _byNameTable that have no associated atom
+	for (const auto &entry: _byNameTable) {
+		const char* name = entry.first;
+		const ld::Atom* atom = _indirectBindingTable[entry.second];
+		if ( (atom == NULL) || (atom->definition() == ld::Atom::definitionProxy) )
+			syms.insert(name);
+	}
+}
+
+
 bool SymbolTable::hasName(const char* name)			
 { 
 	NameToSlot::iterator pos = _byNameTable.find(name);
@@ -813,6 +825,15 @@ SymbolTable::IndirectBindingSlot SymbolTable::findSlotForReferences(const ld::At
 			slot = _indirectBindingTable.size();
 			_pointerToCStringTable[atom] = slot;
 			break;
+		case ld::Section::typeTLVPointers:
+			pos = _threadPointerTable.find(atom);
+			if ( pos != _threadPointerTable.end() ) {
+				*existingAtom = _indirectBindingTable[pos->second];
+				return pos->second;
+			}
+			slot = _indirectBindingTable.size();
+			_threadPointerTable[atom] = slot;
+			break;
 		default:
 			assert(0 && "section type does not support coalescing by references");
 	}
