@@ -315,7 +315,10 @@ ld::relocatable::File* Parser::parseMachOFile(const uint8_t* p, size_t len, cons
 	objOpts.platform			= options.platform;
 	objOpts.subType				= 0;
 	objOpts.srcKind				= ld::relocatable::File::kSourceLTO;
-	
+	objOpts.treateBitcodeAsData = false;
+	objOpts.usingBitcode		= options.bitcodeBundle;
+	objOpts.maxDefaultCommonAlignment = options.maxDefaultCommonAlignment;
+
 	// mach-o parsing is done in-memory, but need path for debug notes
 	const char* path = "/tmp/lto.o";
 	time_t modTime = 0;
@@ -1020,7 +1023,6 @@ bool isObjectFile(const uint8_t* fileContent, uint64_t fileLength, cpu_type_t ar
 	return Parser::validFile(fileContent, fileLength, architecture, subarch);
 }
 
-
 static ld::relocatable::File *parseImpl(
           const uint8_t *fileContent, uint64_t fileLength, const char *path,
           time_t modTime, ld::File::Ordinal ordinal, cpu_type_t architecture,
@@ -1041,6 +1043,12 @@ ld::relocatable::File* parse(const uint8_t* fileContent, uint64_t fileLength,
 								cpu_type_t architecture, cpu_subtype_t subarch, bool logAllFiles,
 								bool verboseOptimizationHints)
 {
+	// do light weight check before acquiring lock
+	if ( fileLength < 4 )
+		return NULL;
+	if ( (fileContent[0] != 0xDE) || (fileContent[1] != 0xC0) || (fileContent[2] != 0x17) || (fileContent[3] != 0x0B) )
+		return NULL;
+
 	// Note: Once lto_module_create_in_local_context() and friends are thread safe
 	// this lock can be removed.
 	Mutex lock;
