@@ -1244,7 +1244,8 @@ static ld::relocatable::File* createReader(const char* path)
 	objOpts.verboseOptimizationHints = true;
 	objOpts.armUsesZeroCostExceptions = true;
 	objOpts.subType				= sPreferredSubArch;
-	objOpts.srcKind				= ld::relocatable::File::kSourceObj;
+	objOpts.treateBitcodeAsData  = false;
+	objOpts.usingBitcode		= true;
 #if 1
 	if ( ! foundFatSlice ) {
 		cpu_type_t archOfObj;
